import pyad.reverse_mode as rev
import numpy as np
import time


def linear_activation(x):
    return x


def softmax_activation(x):
    z = rev.exp(x)
    return z / z.sum(axis=0)


def log_softmax_activation(x):
    return x - rev.log(rev.exp(x).sum(axis=0))


def tanh_activation(x):
    return rev.tanh(x)


def logistic_activation(x):
    return rev.logistic(x)


def relu_activation(x):
    return x * (x >= 0)


def mse_loss(y_pred, y_true):
    return ((y_pred - y_true) ** 2).mean()


def rmse_loss(y_pred, y_true):
    return rev.sqrt(mse_loss(y_pred, y_true))


def nll_loss(y_log_probs, y_true):
    true_idx = y_true.value if isinstance(y_true, rev.Tensor) else y_true
    true_idx = np.array(true_idx, dtype=int)
    return -y_log_probs[true_idx, np.arange(len(true_idx))].mean()


def cross_entropy_loss(y_logits, y_true):
    y_log_probs = log_softmax_activation(y_logits)
    return nll_loss(y_log_probs, y_true)


class NeuralNet:
    """
    A complete neural network class which contains functions for building the
    network, training, evaluation, and prediction.
    """
    activation_funcs = {
        'linear': linear_activation,
        'softmax': softmax_activation,
        'log_softmax': log_softmax_activation,
        'tanh': tanh_activation,
        'logistic': logistic_activation,
        'relu': relu_activation
    }

    loss_functions = {
        'mse': mse_loss,
        'rmse': rmse_loss,
        'nll': nll_loss,
        'cross_entropy': cross_entropy_loss,
    }

    def __init__(self, loss_fn='mse'):
        """
        Initializes an empty NeuralNet with a particular loss function. The
        loss function specified determines what kind of task this neural
        network will be performing.

            `nll` and `cross_entropy` imply classification
            `mse` and `rmse` imply regression

        Parameters
        ----------
        loss_fn: string
            Which loss function should be used. The options are given by
            NeuralNet.loss_functions
        """
        if loss_fn not in self.loss_functions:
            raise ValueError(f'Invalid loss function: {loss_fn}')

        self.layers = []
        self.loss_fn = NeuralNet.loss_functions[loss_fn]
        self.loss_fn_name = loss_fn

    def add_layer(self, in_size, out_size, activation='linear'):
        """
        Adds a layer to the network. The size of the next layer must be
        compatible with the size of the previous layer

        Parameters
        ----------
        in_size: int
            The number of nodes in the previous layer of the network

        out_size: int
            The number of nodes in the next layer of the network

        activation: string
            The name of the activation function which should be used for this
            layer. The options are given by `NeuralNet.activation_funcs`
        """
        if self.layers:
            if self.layers[-1]['out_size'] != in_size:
                raise ValueError('Input size of next layer must match '
                                 'output size of the previous layer.')

        if activation not in self.activation_funcs:
            raise ValueError(f'Invalidation activation function: {activation}')

        # Xavier initialization of weights
        xavier_init_mult = np.sqrt(2 / (out_size * in_size))
        weight_init = np.random.randn(out_size, in_size) * xavier_init_mult
        bias_init = np.zeros((out_size, 1))

        self.layers.append({
            'in_size': in_size,
            'out_size': out_size,
            'weights': rev.Tensor(weight_init),
            'bias': rev.Tensor(bias_init),
            'activation': self.activation_funcs[activation],
        })

    def evaluate(self, x):
        """
        Evaluates the neural network on a given input.

        Parameters
        ----------
        x: np.array
            The input to the network

        Returns
        -------
        rev.Tensor
            The result of running the input through the network
        """
        result = rev.Tensor(x.T)
        for l in self.layers:
            result = l['weights'] @ result + l['bias']
            result = l['activation'](result)
        return result

    def predict(self, x):
        """
        Similar to evaluate(). However, if the loss function is intended
        for a classification task, e.g. nll or cross_entropy, then the
        result will be the actual predicted classes.


        Parameters
        ----------
        x: np.array
            The input to the network

        Returns
        -------
        np.array
            The result of running the input through the network
        """
        res = self.evaluate(x)
        if self.loss_fn_name in ('nll', 'cross_entropy'):
            return res.value.argmax(axis=0)
        return res.value

    def accuracy(self, x, y_true):
        """
        Computes the accuracy of the predictions. This only works for
        classificiation loss functions, i.e. nll or cross_entropy.

        Parameters
        ----------
        x: np.array
            The inputs to the network

        y: np.array
            The true classes

        Returns
        -------
        float
            The accuracy as a float in [0, 1.0]
        """
        if self.loss_fn_name not in ('nll', 'cross_entropy'):
            raise ValueError(f'Cannot compute accuracy for loss_fn {self.loss_fn_name}')

        preds = self.predict(x)
        return (preds == y_true).mean()

    def score(self, x, y):
        """
        Evaluates the loss function on the input.

        Parameters
        ----------
        x: np.array
            The inputs to the network

        y: np.array
            The true values or classes for the given inputs


        Returns
        -------
        float
            The loss
        """
        return self.loss_fn(self.evaluate(x), y)

    def _reset_weight_grads(self):
        """
        Internal helper function for reseting the gradients of all of the
        weights in network
        """
        for l in self.layers:
            l['weights'].reset_grad()
            l['bias'].reset_grad()

    def _update_weights(self, learning_rate, batch_size):
        """
        Internal helper function for performing gradient descent using
        stochastic gradient descent. The gradients are normalized by the
        size of the batch
        """
        learning_rate = learning_rate / batch_size
        for l in self.layers:
            l['weights'] = l['weights'] - l['weights'].grad * learning_rate
            l['bias'] = l['bias'] - l['bias'].grad * learning_rate
        self._reset_weight_grads()

    def train(self, x_train, y_train, x_val, y_val, *,
              batch_size=20, epochs=20, learning_rate=1e-3, verbose=True):
        """
        Iterates through the training data multiple times, and computes the loss
        of the neural network. The weights are then updated by taking their
        gradient w.r.t. the loss using SGD. Periodically logs some output
        about the training progress.

        Parameters
        ----------
        x_train: np.array
            The data which should be used to train the network

        y_train: np.array
            The associated y values for the training data

        x_val: np.array
            The data which should be used to log validation output

        y_val: np.array
            The associated y values for the validation data

        batch_size: int
            How many data samples to use in each SGD batch

        epochs: int
            How many training loops through the data should be executed

        learning_rate: float
            Controls the speed of the SGD update

        verbose: bool
            Whether or not logs should be outputted
        """
        self._reset_weight_grads()

        for epoch in range(epochs):
            try:
                start_time = time.time()
                total_train_loss = 0
                tot_train_correct = 0

                for batch in range(0, len(x_train), batch_size):
                    x_batch = x_train[batch:batch+batch_size]
                    y_batch = y_train[batch:batch+batch_size]
                    y_batch_pred = self.evaluate(x_batch)

                    loss = self.loss_fn(y_batch_pred, y_batch)
                    loss.backward()
                    self._update_weights(learning_rate, len(x_batch))

                    total_train_loss += loss.value * len(x_batch)

                    if self.loss_fn_name in ('nll', 'cross_entropy'):
                        tot_train_correct += (y_batch_pred.value.argmax(0) == y_batch).sum()

                with rev.no_grad():
                    val_preds = self.evaluate(x_val)
                    avg_val_loss = self.loss_fn(val_preds, y_val).value

                if verbose:
                    # Logging some information
                    header = f'Epoch {epoch + 1}'
                    print(header + '\n' + '=' * len(header))
                    print(f'Duration: {time.time() - start_time:.3f} sec')
                    print(f'Avg train loss: {total_train_loss / len(x_train):.3g}')
                    print(f'Avg validation loss: {avg_val_loss:.3g}')

                    if self.loss_fn_name in ('nll', 'cross_entropy'):
                        train_acc = tot_train_correct / len(x_train)
                        val_acc = (val_preds.value.argmax(axis=0) == y_val).mean()
                        print(f'Avg train acc: {train_acc:.3g}')
                        print(f'Avg validation acc: {val_acc:.3g}')

                    print()
            except KeyboardInterrupt:
                print(f'Stopping training after {epoch} epochs...')
                break



import pyad.reverse_mode as rev
import pyad.nn as nn
import numpy as np
import pytest


def test_activations():
    """
    Test the activation functions values. Their derivatives should already be
    correct assuming the underlying primitive operations are correct
    """

    x = rev.Tensor([-1, 2, 3])
    assert nn.linear_activation(x) is x

    softmax = nn.softmax_activation(x)
    assert np.allclose(softmax.value, [0.01321289, 0.26538793, 0.72139918])

    log_softmax = nn.log_softmax_activation(x)
    assert np.allclose(log_softmax.value, [-4.32656264, -1.32656264, -0.32656264])

    tanh = nn.tanh_activation(x)
    assert np.allclose(tanh.value, [-0.76159416,  0.96402758,  0.99505475])

    logistic = nn.logistic_activation(x)
    assert np.allclose(logistic.value, [0.26894142, 0.88079708, 0.95257413])

    relu = nn.relu_activation(x)
    assert np.allclose(relu.value, [0, 2, 3])


def test_regression_loss_functions():
    """
    Test the regression loss functions to make sure their values are correct.
    """
    y_pred = rev.Tensor([-3, 5, 6])
    y_true = rev.Tensor([-5, 1, 7])

    mse = nn.mse_loss(y_pred, y_true)
    assert np.isclose(mse.value, 7.0)

    rmse = nn.rmse_loss(y_pred, y_true)
    assert np.isclose(rmse.value, 7.0 ** 0.5)


def test_classification_loss_functions():
    """
    Test the classification loss functions to make sure their values are correct.
    """
    y_pred = rev.Tensor([[-3, 4, 1], [-2, 5, 3]])
    y_true = rev.Tensor([0, 1, 1])

    nll = nn.nll_loss(y_pred, y_true)
    assert np.isclose(nll.value, -5/3)

    cross_entropy = nn.cross_entropy_loss(y_pred, y_true)
    assert np.isclose(cross_entropy.value, 0.5844837953598061)


def test_regression_NeuralNet():
    """
    Test building and testing a regression NeuralNet
    """

    # try out an invalid loss function
    with pytest.raises(ValueError):
        nn.NeuralNet(loss_fn='blarg')

    net = nn.NeuralNet('mse')

    # try adding an invalid activation
    with pytest.raises(ValueError):
        net.add_layer(10, 10, activation='blarg')

    # add a single layer. hardcode its weights
    net.add_layer(3, 1, activation='linear')
    net.layers[0]['weights'] = rev.Tensor([[1, 2, 3]])
    net.layers[0]['bias'] = rev.Tensor([[1]])

    # try adding an improperly sized layer
    with pytest.raises(ValueError):
        net.add_layer(19, 20, activation='linear')

    # evaluate
    x = np.array([[3, 2, 1], [4, 5, 6]])
    y = np.array([6, 10])
    res = net.evaluate(x)

    assert np.allclose(res.value, [[11, 33]])
    assert net.score(x, y).value == 277

    # predict
    assert np.allclose(net.predict(x), [[11, 33]])

    # accuracy
    with pytest.raises(ValueError):
        net.accuracy(x, y)


def test_classification_NeuralNet():
    """
    Test building and testing a classficiation NeuralNet
    """

    net = nn.NeuralNet(loss_fn='nll')

    # add a singel layer. hardcode its weights
    net.add_layer(3, 2, activation='log_softmax')
    net.layers[0]['weights'] = rev.Tensor([[1, 2, 3], [4, 5, 6]])
    net.layers[0]['bias'] = rev.Tensor([[1], [2]])

    # evaluate
    x = np.array([[3, 2, 1], [4, 5, 6]])
    y = np.array([0, 1])
    res = net.evaluate(x)

    assert np.allclose(res.value, [[-19, -46], [-5.60279645e-09, 0]])
    assert np.isclose(net.score(x, y).value, 9.500000002801398)

    # predict
    assert np.allclose(net.predict(x), [1, 1])

    # accuracy
    assert net.accuracy(x, y) == 0.5


def test_train_NeuralNet():
    """
    Test training a NeuralNet on a simple dataset with 2 features and 2 classes
    """

    np.random.seed(0)
    net = nn.NeuralNet(loss_fn='nll')
    net.add_layer(2, 30, activation='relu')
    net.add_layer(30, 2, activation='log_softmax')

    # generate a classic XOR dataset
    x = np.random.random((250, 2)) - 0.5
    y = (np.sign(x[:, 0]) != np.sign(x[:, 1])).astype(np.int64)

    split = int(len(x) * 0.8)
    x_train, x_val = x[:split], x[split:]
    y_train, y_val = y[:split], y[split:]

    # train it
    net.train(
        x_train, y_train, x_val, y_val,
        batch_size=2, epochs=10, learning_rate=1e-1
    )

    # check that the losses are low and the accuracies are high
    assert net.score(x_train, y_train).value < 0.5
    assert net.score(x_val, y_val).value < 0.5

    assert net.accuracy(x_train, y_train) >= 0.9
    assert net.accuracy(x_val, y_val) >= 0.9



import numpy as np
import pyad.forward_mode as fwd


def newtons_method(f, init, max_iters=1000):
    x = init
    old_res = f(fwd.var('x', x))

    for i in range(max_iters):
        x = x - old_res.value / old_res.d['x']
        res = f(fwd.var('x', x))
        if abs(res.value - old_res.value) < 1e-10:
            break
        old_res = res

    return x


def test_newtons_method():
    # a simple polynomial with three roots
    def f(x):
        return (x - 18) * (x + 50) * (x - 207)

    res = newtons_method(f, 100)
    assert f(res) < 1e-10

    # quintic function with no exact solution
    def f(x):
        return x**5 - x - 1

    res = newtons_method(f, 100)
    assert f(res) < 1e-10

    # a function with infinitely many roots
    def f(x):
        return fwd.sin(fwd.exp(x))

    res = newtons_method(f, 100)
    assert f(res).value < 1e-10


def gauss_newton_method(f, init, max_rounds=100, max_diff=1e-12):
    prev_x = x = init

    for i in range(max_rounds):
        res = f(fwd.var('x', x))
        J = res.d['x']
        x = x - np.linalg.pinv(J.T @ J) @ J.T @ res.value
        if np.linalg.norm(x - prev_x) < max_diff:
            break
        prev_x = x
    return x


def test_gauss_newton():
    """
    Test a version of Gauss newton for vector functions
    """
    def norm(x, d=2):
        return sum(v**d for v in x) ** (1/d)

    def f(d):
        x = norm(d, 4) - 1
        y = norm([3*d[0] - 1*d[1], 1*d[0] + 0*d[1]], 4) - 1
        return fwd.stack([x, y])

    init_points = [
        [1, 1], [-1, 1], [1, -1], [-1, -1]
    ]

    for init in init_points:
        sol = gauss_newton_method(f, init)
        assert f(sol).norm().value < 1e-10



import numpy as np

# we use this to get rid of disruptive runtime warnings
runtime_warning_filter = np.testing.suppress_warnings()
runtime_warning_filter.filter(RuntimeWarning)


class Tensor:
    """
    Class for automatic differentiation reverse mode
    """
    _gradients_disabled = False

    def __init__(self, value):
        value = value.value if isinstance(value, Tensor) else value
        self.value = np.array(value, dtype=np.float64)
        self.children = []
        self.grad_value = None

        if len(self.value.shape) > 2:
            raise ValueError('Can only support 0-D, 1-D, and 2-D Tensors')

    def backward(self):
        """
        A function that seeds in the derivative of a function with respect to
        itself, i.e. df/df = 1.

        backward() can only be called on a scalar Tensor. pyad.reverse_mode
        does not support computing the gradients of vector outputs
        """
        if self.value.shape != ():
            raise ValueError('Cannot call .backward() on a non-scalar')
        self.grad_value = np.array(1.0)

    def add_child(self, weight, output_tensor):
        """
        Adds a node to the computation graph.

        Parameters
        ----------
        weight: callable | np.array
            Specifies how the gradient w.r.t. the output should be
            calculated

        output_tensor: Tensor
            A Tensor that is dependent on this node
        """
        if not Tensor._gradients_disabled:
            self.children.append((weight, output_tensor))

    @property
    def grad(self):
        """
        A function that computes the gradient value using the chain rule
        """
        if self.grad_value is None:
            self.grad_value = np.array(0.0)
            for weight, node in self.children:
                if callable(weight):
                    update = weight(node.grad)
                else:
                    update = weight * node.grad

                # fix any issues with broadcasting by summing over
                # the broadcasted dimensions
                while (update.shape != ()
                       and update.shape != self.shape
                       and update.shape + (1,) != self.shape):
                    update = update.sum(axis=-1)

                if update.shape + (1,) == self.shape:
                    update = np.expand_dims(update, axis=-1)

                # add the corrected derivative to the gradient
                self.grad_value = self.grad_value + update

        return self.grad_value

    # Comparisons work like they do in numpy. Derivative information is ignored
    # numpy arrays are returned by comparisons
    def __lt__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        return self.value < other.value

    def __gt__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        return self.value > other.value

    def __le__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        return self.value <= other.value

    def __ge__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        return self.value >= other.value

    def __eq__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        return self.value == other.value

    def __ne__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        return self.value != other.value

    def __bool__(self):
        return bool(self.value)

    def __repr__(self):
        return f'Tensor({self.value}, D({self.grad_value}))'

    def reset_grad(self):
        """
        Resets the gradient of the Tensor.
        """
        self.grad_value = None
        self.children = []

    @property
    def shape(self):
        """
        Returns the shape of the underlying np.array
        """
        return self.value.shape

    def sum(self, axis=None):
        """
        Returns the sum of all of the elements in the Tensor. Optionally,
        along a particular axis.

        Parameters
        ----------
        axis: int | None
            Which axis should the sum() be taken over

        Returns
        -------
        Tensor
            A new Tensor that has had the sum() operation applied
        """
        def sum_backward(out_grad):
            return np.broadcast_to(out_grad, self.shape)

        z = Tensor(np.sum(self.value, axis=axis))
        self.add_child(sum_backward, z)
        return z

    def mean(self):
        """
        Returns the arithmetic mean of all of the elements in the Tensor
        """
        return self.sum() / self.value.size

    def prod(self):
        """
        Returns the product of all of the elements in the Tensor
        """
        if self.shape == ():
            return self

        forward_prod = np.cumprod(self.value.flatten())
        backward_prod = np.cumprod(self.value.flatten()[::-1])[::-1]
        result = np.ones(len(forward_prod))

        for i in range(len(forward_prod)):
            if i != 0:
                result[i] *= forward_prod[i - 1]
            if i != len(forward_prod) - 1:
                result[i] *= backward_prod[i + 1]

        z = Tensor(np.prod(self.value))
        self.add_child(result.reshape(self.shape), z)
        return z

    def __neg__(self):
        return self.__mul__(-1)

    def __add__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        z = Tensor(self.value + other.value)
        self.add_child(1.0, z)
        other.add_child(1.0, z)
        return z

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return (-self) + other

    def __mul__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        z = Tensor(self.value * other.value)
        self.add_child(other.value, z)
        other.add_child(self.value, z)
        return z

    def __rmul__(self, other):
        return self * other

    @runtime_warning_filter
    def __truediv__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        z = Tensor(self.value / other.value)
        self.add_child(1 / other.value, z)
        other.add_child(-self.value / other.value**2, z)
        return z

    def __rtruediv__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        return other / self

    @runtime_warning_filter
    def __pow__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        z = Tensor(self.value ** other.value)
        self.add_child(other.value * self.value**(other.value - 1), z)
        other.add_child(self.value**other.value * np.log(self.value), z)
        return z

    def __rpow__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        return other ** self

    def _to_2d(self, x):
        """
        A helper function for reshaping numpy arrays as 2D matrices
        """
        return x.reshape(len(x), -1)

    def __matmul__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)

        def mm_backward_self(out_grad):
            other_mat = self._to_2d(other.value)
            out_grad_mat = self._to_2d(out_grad)
            return (out_grad_mat @ other_mat.T).reshape(self.value.shape)

        def mm_backward_other(out_grad):
            self_mat = self._to_2d(self.value)
            out_grad_mat = self._to_2d(out_grad)
            return (self_mat.T @ out_grad_mat).reshape(other.value.shape)

        z = Tensor(self.value @ other.value)
        self.add_child(mm_backward_self, z)
        other.add_child(mm_backward_other, z)
        return z

    def __rmatmul__(self, other):
        other = other if isinstance(other, Tensor) else Tensor(other)
        return other @ self

    def __getitem__(self, idx):
        def getitem_backward(out_grad):
            res = np.zeros(self.shape)
            res[idx] = out_grad
            return res

        z = Tensor(self.value[idx])
        self.add_child(getitem_backward, z)
        return z

    def __setitem__(self, idx, other):
        def setitem_backward(out_grad):
            return out_grad[idx]

        other = other if isinstance(other, Tensor) else Tensor(other)
        other.add_child(setitem_backward, self)
        self.value[idx] = other.value
        return self


class no_grad:
    """
    This is a context manager. Within the no_grad() context, reverse_mode
    operations will not add onto the computation graph.
    """
    def __enter__(self):
        Tensor._gradients_disabled = True

    def __exit__(self, typ, val, traceback):
        Tensor._gradients_disabled = False


# Elementary functions
@runtime_warning_filter
def _elementary_op(obj, fn, deriv_fn):
    """
    A generic framework to allow for the chain rule of other elementary
    functions taken from the numpy module.
    Parameters
    ----------
    obj : Scalar or Tensor object which the elementary function is being
        differentiated at
    fn : np.function
        The elementary function from the numpy module
    deriv_fun:  np.function
        The prespecified derivative of the given numpy elementary function
    Returns
    -------
    Tensor: class
        Tensor object which contains the resulting value and result from the
        chain rule (new derivative)
    """
    obj = obj if isinstance(obj, Tensor) else Tensor(obj)
    z = Tensor(fn(obj.value))
    obj.add_child(deriv_fn(obj.value), z)
    return z


def sin(x):
    return _elementary_op(x, np.sin, np.cos)


def cos(x):
    return _elementary_op(x, np.cos, lambda x: -np.sin(x))


def tan(x):
    return _elementary_op(x, np.tan, lambda x: 1 / (np.cos(x) ** 2))


def arcsin(x):
    return _elementary_op(x, np.arcsin, lambda x: 1 / np.sqrt(1 - x ** 2))


def arccos(x):
    return _elementary_op(x, np.arccos, lambda x: -1 / np.sqrt(1 - x ** 2))


def arctan(x):
    return _elementary_op(x, np.arctan, lambda x: 1 / (1 + x ** 2))


def sinh(x):
    return _elementary_op(x, np.sinh, np.cosh)


def cosh(x):
    return _elementary_op(x, np.cosh, np.sinh)


def tanh(x):
    return _elementary_op(x, np.tanh, lambda x: 1 / (np.cosh(x) ** 2))


def abs(x):
    return _elementary_op(x, np.abs, np.sign)


def exp(x):
    return _elementary_op(x, np.exp, np.exp)


def logistic(x):
    return 1 / (1 + exp(-x))


def log(x, base=np.e):
    if base == np.e:
        return _elementary_op(x, np.log, lambda x: 1 / x)
    return log(x) / log(base)


def log2(x):
    return log(x, base=2)


def log10(x):
    return log(x, base=10)


def sqrt(x):
    return _elementary_op(x, np.sqrt, lambda x: 1 / (2 * np.sqrt(x)))


def cbrt(x):
    return _elementary_op(x, np.cbrt, lambda x: 1 / (3 * x ** (2/3)))



import reverse_mode as rev
import matplotlib.pyplot as plt

x = rev.Tensor(3)
y = rev.Tensor(2)

f = x * y**3 + x / 3

f.backward()

graph_obj = rev.rev_graph()
plot = graph_obj.plot_graph([x, y])

plt.show()


import reverse_mode as rev
import numpy as np
import matplotlib.pyplot as plt

x = rev.Tensor(0.5)
y = rev.Tensor(4.2)
z = rev.Tensor(3)
f = x * y**3 + rev.sin(x) - rev.logistic(z)

#set df seed
f.backward()

rev_g = rev.rev_graph()
plot = rev_g.plot_graph([x,y,z])

plt.show()


import numpy as np
import math
import pyad.forward_mode as fwd
import pytest

################################
#       End to End Tests       #
################################

def test_case_base_trigonometric():
    """
    Try some end to end testing using a complex trigonometric function of sin, cos and tan
    """
    x = fwd.var('x', 1)
    y = fwd.var('y', 0.5)
    z = fwd.var('z', 2)
    function = fwd.cos(x) + 3*(fwd.sin(y)**2) * fwd.cos(z) + fwd.tan(x)

    # Calculated from Wolfram Alpha
    true_value = 1.8107574187515
    true_x_deriv = 2.5840478360068
    true_y_deriv = -1.0505264651220
    true_z_deriv = -0.6270028955876

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(function.d['x'], true_x_deriv, rel_tol=1e-12)
    assert math.isclose(function.d['y'], true_y_deriv, rel_tol=1e-12)
    assert math.isclose(function.d['z'], true_z_deriv, rel_tol=1e-12)


def test_case_base_inversetrig():
    """
    Testing inverse trig functions
    """
    x = fwd.var('x', 0.25)
    y = fwd.var('y', 0.5)
    z = fwd.var('z', 3)
    function = fwd.arcsin(x)**3 + fwd.arccos(y)**2 - z*fwd.arctan(z)

    # Calculated using numpy
    true_value = -2.634381651043423
    true_x_deriv = 0.1978236588118186
    true_y_deriv = -2.4183991523122903
    true_z_deriv = -1.5490457723982545

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(function.d['x'], true_x_deriv, rel_tol=1e-12)
    assert math.isclose(function.d['y'], true_y_deriv, rel_tol=1e-12)
    assert math.isclose(function.d['z'], true_z_deriv, rel_tol=1e-12)


def test_case_base_hyperbolic():
    """
    Testing hyperbolic functions
    """
    x = fwd.var('x', 0.5)
    y = fwd.var('y', 1)
    z = fwd.var('z', 3)
    function = 2*fwd.sinh(x)*x + 3*fwd.cosh(y) + fwd.tanh(z)

    # Calculated using numpy
    true_value = 6.145391963626209
    true_x_deriv = 2.1698165761938757
    true_y_deriv = 3.525603580931404
    true_z_deriv = 0.009866037165440192

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(function.d['x'], true_x_deriv, rel_tol=1e-12)
    assert math.isclose(function.d['y'], true_y_deriv, rel_tol=1e-12)
    assert math.isclose(function.d['z'], true_z_deriv, rel_tol=1e-12)


def test_case_base_exp():
    """
    Testing exponential function
    """
    x = fwd.var('x', 2)
    function = fwd.exp(x) + fwd.exp(-x)

    # Calculated using numpy
    true_value = 7.524391382167263
    true_x_deriv = 7.253720815694038

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(function.d['x'], true_x_deriv, rel_tol=1e-12)


def test_case_base_log():
    """
    Testing logarithm function
    """
    x = fwd.var('x', 3)
    function = fwd.log(x)**2

    # Calculated using numpy
    true_value = 1.206948960812582
    true_x_deriv = 0.7324081924454066

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(function.d['x'], true_x_deriv, rel_tol=1e-12)


def test_log2():
    """
    Testing logarithm base 2 function
    """
    x = fwd.var('x', 3)
    function = fwd.log2(x) ** 2

    # Calculated using numpy
    true_value = np.log2(3) ** 2
    true_x_deriv = 2 * np.log2(3) / 3 / np.log(2)

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(function.d['x'], true_x_deriv, rel_tol=1e-12)


def test_log10():
    """
    Testing logarithm base 10 function
    """
    x = fwd.var('x', 3)
    function = fwd.log10(x) ** 2

    # Calculated using numpy
    true_value = np.log10(3) ** 2
    true_x_deriv = 2 * np.log10(3) / 3 / np.log(10)

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(function.d['x'], true_x_deriv, rel_tol=1e-12)


def test_case_base_root():
    """
    Testing square root and cubic root functions
    """
    x = fwd.var('x', 4)
    y = fwd.var('y', 8)
    function = fwd.sqrt(x) + fwd.cbrt(y)

    # Calculated using numpy
    true_value = 4
    true_x_deriv = 0.25
    true_y_deriv = 0.0833333333333333

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(function.d['x'], true_x_deriv, rel_tol=1e-12)
    assert math.isclose(function.d['y'], true_y_deriv, rel_tol=1e-12)


def test_case_base_abs():
    """
    Testing absolute value functin
    """
    x = fwd.var('x', -2)
    function = fwd.abs(x)

    true_value = 2
    true_x_deriv = -1

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(function.d['x'], true_x_deriv, rel_tol=1e-12)


##########################
#       Unit Tests       #
##########################


#### Tests on the Tensor Class Object ####

def test_tensor_from_another_tensor():
    t1 = fwd.Tensor(5, d=fwd.MultivariateDerivative({'a': 2, 'b': 3}))
    t2 = fwd.Tensor(t1)
    assert t1.value == t2.value
    assert t1.d.variables == t2.d.variables


def test_get_value_and_deriv_return_tensor():
    """
    Get the value and derivative of a tensor
    """
    test_tensor = fwd.Tensor(5)
    output_tuple = test_tensor.get_value_and_deriv(test_tensor)

    assert type(output_tuple) == tuple
    assert output_tuple[0] == np.array(5)
    assert type(output_tuple[1]) == fwd.MultivariateDerivative
    assert output_tuple[1].variables == {}


def test_get_value_and_deriv_return_constant():
    """
    Get the value and derivative of a constant
    """
    test_tensor = fwd.Tensor(5)
    output_tuple = test_tensor.get_value_and_deriv(1)

    assert type(output_tuple) == tuple
    assert output_tuple[0] == 1
    assert type(output_tuple[1]) == fwd.MultivariateDerivative
    assert output_tuple[1].variables == {}


def test_get_value_and_deriv_return_Variable():
    """
    Get the value and derivative of a variable
    """
    test_tensor = fwd.Tensor(5)
    test_variable = fwd.Variable('x', 2)
    output_tuple = test_tensor.get_value_and_deriv(test_variable)

    assert type(output_tuple) == tuple
    assert output_tuple[0] == np.array(2)
    assert type(output_tuple[1]) == fwd.MultivariateDerivative
    assert output_tuple[1].variables == {'x':1}


def test_norm_tensor_single_value():
    """
    Ensure the norm function correctly calculates euclidian norm
    """
    test_tensor = fwd.Variable('x', 2) + fwd.Variable('y', 3)
    norm = test_tensor.norm()

    assert norm.value == 5
    assert norm.d['x'] == 1
    assert norm.d['y'] == 1


def test_norm_tensor_single_vector():
    """
    Ensure the norm function correctly calculates euclidian norm
    """
    test_tensor = fwd.Tensor([2, 2, 1]) * fwd.Variable('x', 4)
    norm = test_tensor.norm()

    assert norm.value == 12
    assert norm.d['x'] == 3


def test_sum():
    """
    Tests the sum function
    """

    # test sum of a vector
    t = fwd.Tensor([1, 2, 3]) * fwd.var('x', 1) + fwd.var('y', 3)
    res = t.sum()

    assert res.value == 15
    assert res.d['x'] == 6
    assert res.d['y'] == 3

    # test sum of a scalar
    assert res.shape == ()

    res = res.sum()
    assert res.value == 15
    assert res.d['x'] == 6
    assert res.d['y'] == 3


def test_repr():
    """
    Ensure repr correctly prints the Tensor class
    """
    test_tensor = fwd.Tensor(3)
    test_tensor2 = fwd.Tensor(3, d=fwd.MultivariateDerivative({'x':5}))

    assert test_tensor.__repr__() == 'Tensor(3, D())'
    assert test_tensor2.__repr__() == 'Tensor(3, D(x=5))'


def test_neg():
    """
    Ensure negating works on both value and derivatives
    """
    test_tensor = fwd.Tensor(3, d=fwd.MultivariateDerivative({'x':5}))
    neg_tensor = test_tensor.__neg__()

    assert neg_tensor.value == -3
    assert neg_tensor.d.variables == {'x':-5}


def test_add():
    """
    Add two tensors together to get a new tensor
    """
    test_tensor1 = fwd.Tensor(3, d=fwd.MultivariateDerivative({'x':5, 'y':10}))
    test_tensor2 = fwd.Tensor(5, d=fwd.MultivariateDerivative({'x':3, 'y':30}))

    new_tensor = test_tensor1 + test_tensor2

    assert type(new_tensor) == fwd.Tensor
    assert new_tensor.value == 8
    assert new_tensor.d.variables == {'x':8, 'y':40}


def test_radd():
    """
    Add a non-tensor and a tensor
    """
    t1 = fwd.Tensor(3, d=fwd.MultivariateDerivative({'x': 5, 'y': 10}))
    res = 5 + t1

    assert isinstance(res, fwd.Tensor)
    assert res.value == 8
    assert res.d.variables == {'x': 5, 'y': 10}


def test_rsub():
    """
    Subtract a non-tensor and a tensor
    """
    t1 = fwd.Tensor(3, d=fwd.MultivariateDerivative({'x': 5, 'y': 10}))
    res = 5 - t1

    assert isinstance(res, fwd.Tensor)
    assert res.value == 2
    assert res.d.variables == {'x': -5, 'y': -10}


def test_rtruedriv():
    """
    Divide a non-tensor and a tensor
    """
    t1 = fwd.Tensor(3, d=fwd.MultivariateDerivative({'x': 5, 'y': 10}))
    res = -5 / t1

    assert isinstance(res, fwd.Tensor)
    assert res.value == -5/3
    assert res.d.variables == {'x': 25/9, 'y': 50/9}


def test_rpow():
    """
    Take a non-tensor to the power of a tensor
    """
    t1 = fwd.Tensor(3, d=fwd.MultivariateDerivative({'x': 5, 'y': 10}))
    res = 4 ** t1

    assert isinstance(res, fwd.Tensor)
    assert res.value == 64
    assert res.d.variables == {'x': 320 * math.log(4), 'y': 640 * math.log(4)}


def test_logistic():
    """
    Tests the logistic function
    """
    t1 = fwd.Tensor(3, d=fwd.MultivariateDerivative({'x': 5, 'y': 10}))
    res = fwd.logistic(t1)

    expected_value = 1 / (1 + np.exp(-3))
    expected_deriv = np.exp(3) / (np.exp(3) + 1) ** 2
    assert math.isclose(res.value, expected_value, rel_tol=1e-12)
    assert math.isclose(res.d['x'], 5 * expected_deriv, rel_tol=1e-12)
    assert math.isclose(res.d['y'], 10 * expected_deriv, rel_tol=1e-12)


#### Tests on the MultivariateDerivative Class ####

def test_repr_mvd():
    """
    Ensure multivariatederivates print correctly
    """
    mvd = fwd.MultivariateDerivative({'x':10})

    assert mvd.__repr__() == 'D(x=10)'


def test_copy_mvd():
    """
    Ensure copy is a new MVD object
    """
    mvd = fwd.MultivariateDerivative({'x':10})
    mvd2 = mvd.copy()

    assert mvd2.variables == {'x':10}
    assert mvd2 != mvd


def test_getitem_mvd():
    """
    Ensure get item returns correctly
    """
    mvd = fwd.MultivariateDerivative({'x':10, 'y':3, 'z':1})

    assert mvd.__getitem__('x') == 10
    assert mvd.__getitem__('y') == 3
    assert mvd.__getitem__('z') == 1


def test_add_mvd():
    """
    Test the add method for the multivariatederivative object
    """
    mvd1 = fwd.MultivariateDerivative({'x':10, 'y':3, 'z':1})
    mvd2 = fwd.MultivariateDerivative({'x':10, 'y':3, 'a':1})

    mvd3 = mvd1 + mvd2

    assert mvd3.variables == {'x':20, 'y':6, 'z':1, 'a':1}


def test_mul_mvd():
    """
    Test the multiply method for the multivariatederivative object
    """
    mvd1 = fwd.MultivariateDerivative({'x':10, 'y':3, 'z':1})

    mvd2 = mvd1.mul(10)

    assert mvd2.variables == {'x':100, 'y':30, 'z':10}


#### Tests on non-class functions ####

def test_tensor_function():
    """
    Check that the tensor wrapper around Tensor correctly returns a Tensor object
    """
    test_tensor1 = fwd.tensor(5)

    assert type(test_tensor1) == fwd.Tensor
    assert test_tensor1.value == np.array(5)
    assert type(test_tensor1.d) == fwd.MultivariateDerivative
    assert test_tensor1.d.variables == {}


def test_var_function():
    """
    Check that the var wrapper around Variable correctly returns a Variable
    """
    test_var1 = fwd.var('x', 5)

    assert type(test_var1) == fwd.Variable
    assert test_var1.value == np.array(5)
    assert type(test_var1.d) == fwd.MultivariateDerivative
    assert test_var1.d.variables == {'x':1}


#### Tests on the Tensor Class Object ####

def test_comparisons():
    """
    Check that comparison operators work for Tensors
    """
    t1 = fwd.Tensor([1, 12, 3, 4, 5])
    t2 = fwd.Tensor([4, 2, 3, 1, 10])

    # test against another Tensor
    assert np.array_equal(t1 < t2, np.array([1, 0, 0, 0, 1]))
    assert np.array_equal(t1 > t2, np.array([0, 1, 0, 1, 0]))
    assert np.array_equal(t1 <= t2, np.array([1, 0, 1, 0, 1]))
    assert np.array_equal(t1 >= t2, np.array([0, 1, 1, 1, 0]))
    assert np.array_equal(t1 == t2, np.array([0, 0, 1, 0, 0]))
    assert np.array_equal(t1 != t2, np.array([1, 1, 0, 1, 1]))

    # test against a non-Tensor
    assert np.array_equal(t1 < 3, np.array([1, 0, 0, 0, 0]))
    assert np.array_equal(t1 > 3, np.array([0, 1, 0, 1, 1]))
    assert np.array_equal(t1 <= 3, np.array([1, 0, 1, 0, 0]))
    assert np.array_equal(t1 >= 3, np.array([0, 1, 1, 1, 1]))
    assert np.array_equal(t1 == 3, np.array([0, 0, 1, 0, 0]))
    assert np.array_equal(t1 != 3, np.array([1, 1, 0, 1, 1]))


def test_any_all():
    """
    Tests the any() and all() methods
    """
    t1 = fwd.Tensor([0, 0, 0])
    assert t1.any() is False
    assert t1.all() is False

    t2 = fwd.Tensor([1, 2, 3])
    assert t2.any() is True
    assert t2.all() is True

    t3 = fwd.Tensor([0, 1, 2])
    assert t3.any() is True
    assert t3.all() is False

    t4 = fwd.Tensor(1)
    assert t4.any() is True
    assert t4.all() is True

    t5 = fwd.Tensor([])
    assert t5.any() is False
    assert t5.all() is True


def test_bool():
    """
    Check that you should only be able to call bool() on a singleton Tensor
    """
    t0 = fwd.Tensor(0)
    t1 = fwd.Tensor([1])
    t2 = fwd.Tensor([1, 2, 3, 4])

    assert bool(t0) is False
    assert bool(t1) is True

    with pytest.raises(ValueError):
        bool(t2)


### Tests for vector operations ###
def test_simple_vector_arithmetic():
    """
    Tests simple arithmetic (add/sub/mul/div) for vectors
    """
    v1 = fwd.var('x', 5)
    v2 = fwd.var('y', 10)

    t1 = fwd.Tensor([1, 2, 3])
    t2 = fwd.Tensor([5, 6, 7])
    res = t1*v1 - t2*v2

    assert np.array_equal(res.value, np.array([-45, -50, -55]))
    assert np.array_equal(res.d['x'], np.array([1, 2, 3]))
    assert np.array_equal(res.d['y'], np.array([-5, -6, -7]))

    res = 5 + (res * res * res) / res
    assert np.array_equal(res.value, np.array([2030, 2505, 3030]))
    assert np.array_equal(res.d['x'], np.array([-90, -200, -330]))
    assert np.array_equal(res.d['y'], np.array([450, 600, 770]))


def test_scalar_var_getitem():
    """
    Tests indexing into a Tensor with scalar variables
    """

    v1 = fwd.var('x', 5)
    v2 = fwd.var('y', 10)

    t1 = fwd.Tensor([1, 2, 3])
    t2 = fwd.Tensor([5, 6, 7])

    res = t1*v1 - t2*v2
    res = 5 + (res * res * res) / res

    assert isinstance(res[0], fwd.Tensor)
    assert res[0].value == 2030
    assert res[0].d['x'] == -90
    assert res[0].d['y'] == 450

    assert isinstance(res[1], fwd.Tensor)
    assert res[1].value == 2505
    assert res[1].d['x'] == -200
    assert res[1].d['y'] == 600

    assert isinstance(res[2], fwd.Tensor)
    assert res[2].value == 3030
    assert res[2].d['x'] == -330
    assert res[2].d['y'] == 770

    assert isinstance(res[0:2], fwd.Tensor)
    assert np.array_equal(res[0:2].value, np.array([2030, 2505]))
    assert np.array_equal(res[0:2].d['x'], np.array([-90, -200]))
    assert np.array_equal(res[0:2].d['y'], np.array([450, 600]))

    assert isinstance(res[[0, 2]], fwd.Tensor)
    assert np.array_equal(res[[0, 2]].value, np.array([2030, 3030]))
    assert np.array_equal(res[[0, 2]].d['x'], np.array([-90, -330]))
    assert np.array_equal(res[[0, 2]].d['y'], np.array([450, 770]))


def test_scalar_var_setitem():
    """
    Tests setting indexes in a Tensor with scalar variables
    """
    v1 = fwd.var('x', 5)
    v2 = fwd.var('y', 10)
    v3 = fwd.var('z', 5)

    t = fwd.Tensor([1, 2, 3, 4]) * v1 + v2 - 2 * v3

    assert np.array_equal(t.value, np.array([5, 10, 15, 20]))
    assert np.array_equal(t.d['x'], np.array([1, 2, 3, 4]))
    assert np.array_equal(t.d['y'], np.array([1, 1, 1, 1]))
    assert np.array_equal(t.d['z'], np.array([-2, -2, -2, -2]))

    t[0] = t[1]
    assert np.array_equal(t.value, np.array([10, 10, 15, 20]))
    assert np.array_equal(t.d['x'], np.array([2, 2, 3, 4]))
    assert np.array_equal(t.d['y'], np.array([1, 1, 1, 1]))
    assert np.array_equal(t.d['z'], np.array([-2, -2, -2, -2]))

    t[1:3] = t[3]
    assert np.array_equal(t.value, np.array([10, 20, 20, 20]))
    assert np.array_equal(t.d['x'], np.array([2, 4, 4, 4]))
    assert np.array_equal(t.d['y'], np.array([1, 1, 1, 1]))
    assert np.array_equal(t.d['z'], np.array([-2, -2, -2, -2]))

    # adding a variable that previously wasn't there
    t[0] = fwd.var('q', 100)
    assert np.array_equal(t.value, np.array([100, 20, 20, 20]))
    assert np.array_equal(t.d['x'], np.array([0, 4, 4, 4]))
    assert np.array_equal(t.d['y'], np.array([0, 1, 1, 1]))
    assert np.array_equal(t.d['z'], np.array([0, -2, -2, -2]))
    assert np.array_equal(t.d['q'], np.array([1, 0, 0, 0]))


def test_vector_var():
    """
    Basic arithmetic tests for vector variables
    """
    v1 = fwd.var('x', [1, 2, 3])
    v2 = fwd.var('y', [2, 3, 4])

    assert np.array_equal(v1.d['x'], np.eye(3))
    assert np.array_equal(v2.d['y'], np.eye(3))

    res = (v1 + 1) * (v2 + 1)
    assert np.array_equal(res.d['x'], np.diag([3, 4, 5]))
    assert np.array_equal(res.d['y'], np.diag([2, 3, 4]))

    # try out invalid variable declarations
    with pytest.raises(ValueError):
        fwd.var('x', [])

    with pytest.raises(ValueError):
        fwd.var('x', np.ones((2, 2)))


def test_vector_var_getitem():
    """
    Tests getting indexes in a Tensor with vector variables
    """
    v1 = fwd.var('x', [1, 2, 3])
    v2 = fwd.var('y', [2, 3, 4])
    res = 6 * (v1 + 1) / (v2 - 1)

    assert np.array_equal(res.value, np.array([12, 9, 8]))
    assert np.array_equal(res.d['x'], np.diag([6, 3, 2]))
    assert np.array_equal(res.d['y'], np.diag([-12, -9/2, -8/3]))

    assert np.array_equal(res[0].value, 12)
    assert np.array_equal(res[0].d['x'], np.array([6, 0, 0]))
    assert np.array_equal(res[0].d['y'], np.array([-12, 0, 0]))

    assert np.array_equal(res[1].value, 9)
    assert np.array_equal(res[1].d['x'], np.array([0, 3, 0]))
    assert np.array_equal(res[1].d['y'], np.array([0, -9/2, 0]))

    assert np.array_equal(res[2].value, 8)
    assert np.array_equal(res[2].d['x'], np.array([0, 0, 2]))
    assert np.array_equal(res[2].d['y'], np.array([0, 0, -8/3]))


def test_vector_var_setitem():
    """
    Tests setting indexes in a Tensor with vector variables
    """
    v1 = fwd.var('x', [1, 2, 3])
    v2 = fwd.var('y', [2, 3, 4])
    res = 2 * v1 * v2

    res[0] += res[1] + res[2] / 2
    assert np.array_equal(res.value, [28, 12, 24])
    assert np.array_equal(res.d['x'], [[4, 6, 4], [0, 6, 0], [0, 0, 8]])
    assert np.array_equal(res.d['y'], [[2, 4, 3], [0, 4, 0], [0, 0, 6]])


def test_shape_and_len():
    """
    Tests len() and .shape of a Tensor
    """
    v1 = fwd.Tensor(0)
    assert v1.shape == ()
    with pytest.raises(TypeError):
        len(v1)

    v2 = fwd.Tensor([0])
    assert v2.shape == (1,)
    assert len(v2) == 1

    v3 = fwd.Tensor([0, 1, 2, 3])
    assert v3.shape == (4,)
    assert len(v3) == 4

    v4 = fwd.Tensor([[1, 2, 3], [4, 5, 6]])
    assert v4.shape == (2, 3)
    assert len(v4) == 2


def test_iter():
    """
    Tests iterating over a Tensor
    """
    v1 = fwd.Tensor(0)
    with pytest.raises(TypeError):
        iter(v1)

    v2 = fwd.Tensor([0, 1], d=fwd.MultivariateDerivative({'x': [1, 2], 'y': [3, 4]}))
    res = []
    for x in v2:
        res.append(x)

    assert res[0].value == 0
    assert res[0].d['x'] == 1
    assert res[0].d['y'] == 3

    assert res[1].value == 1
    assert res[1].d['x'] == 2
    assert res[1].d['y'] == 4


def test_stack():
    """
    Tests for stacking Tensors
    """

    # stack with scalar variables
    t1 = fwd.Tensor([1, 2, 3]) * fwd.var('x', 2)
    t2 = fwd.Tensor([4, 5, 6]) * fwd.var('y', 3)

    res = fwd.stack([t1, t2])
    assert np.array_equal(res.value, np.array([[2, 4, 6], [12, 15, 18]]))
    assert np.array_equal(res.d['x'], np.array([[1, 2, 3], [0, 0, 0]]))
    assert np.array_equal(res.d['y'], np.array([[0, 0, 0], [4, 5, 6]]))

    # stack with vector variables
    t1 = fwd.Tensor([2, 3, 4]) * fwd.var('x', [1, 2, 3])
    t2 = fwd.Tensor([5, 6, 7]) * fwd.var('y', [4, 5, 6])

    res = fwd.stack([t1.sum(), t2.sum()])
    assert np.array_equal(res.value, np.array([20, 92]))
    assert np.array_equal(res.d['x'], [[2, 3, 4], [0, 0, 0]])
    assert np.array_equal(res.d['y'], [[0, 0, 0], [5, 6, 7]])

    res = res + fwd.var('a', 1)
    res = res * fwd.var('b', [3, 4])
    assert np.array_equal(res.d['a'], [3, 4])
    assert np.array_equal(res.d['b'], [[21, 0], [0, 93]])

    # test stack of an empty list
    with pytest.raises(ValueError):
        fwd.stack([])

    # test stack with inconsistent shapes
    with pytest.raises(ValueError):
        fwd.stack([fwd.var('x', 1), fwd.var('x', [1, 2, 3])])



import numpy as np
import math
import pyad.reverse_mode as rev
import pytest


################################
#       End to End Tests       #
################################
def test_reverse_by_example():
    x = rev.Tensor(0.5)
    y = rev.Tensor(4.2)
    z = rev.Tensor(3)
    f = x * y**3 + rev.sin(x) - rev.logistic(z)

    # set df seed
    f.backward()

    assert abs(f.value - (0.5*4.2**3+np.sin(0.5) - 1/(1+np.exp(-3)))) <= 1e-15
    assert abs(x.grad - (4.2**3 + np.cos(0.5))) <= 1e-15
    assert abs(y.grad - (3*0.5*4.2**2)) <= 1e-15
    assert abs(z.grad - (-np.exp(-3)/(1+np.exp(-3))**2)) <= 1e-15


def test_case_base_trigonometric():
    """
    Try some end to end testing using a complex trigonometric function of sin, cos and tan
    """
    x = rev.Tensor(1)
    y = rev.Tensor(0.5)
    z = rev.Tensor(2)
    function = rev.cos(x) + 3*(rev.sin(y)**2) * rev.cos(z) + rev.tan(x)

    function.backward()

    # Calculated from Wolfram Alpha
    true_value = 1.8107574187515
    true_x_deriv = 2.5840478360068
    true_y_deriv = -1.0505264651220
    true_z_deriv = -0.6270028955876

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(x.grad, true_x_deriv, rel_tol=1e-12)
    assert math.isclose(y.grad, true_y_deriv, rel_tol=1e-12)
    assert math.isclose(z.grad, true_z_deriv, rel_tol=1e-12)


def test_case_base_inversetrig():
    """
    Testing inverse trig functions
    """
    x = rev.Tensor(0.25)
    y = rev.Tensor(0.5)
    z = rev.Tensor(3)
    function = rev.arcsin(x)**3 + rev.arccos(y)**2 - z*rev.arctan(z)

    function.backward()

    # Calculated using numpy
    true_value = -2.634381651043423
    true_x_deriv = 0.1978236588118186
    true_y_deriv = -2.4183991523122903
    true_z_deriv = -1.5490457723982545

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(x.grad, true_x_deriv, rel_tol=1e-12)
    assert math.isclose(y.grad, true_y_deriv, rel_tol=1e-12)
    assert math.isclose(z.grad, true_z_deriv, rel_tol=1e-12)


def test_case_base_hyperbolic():
    """
    Testing hyperbolic functions
    """
    x = rev.Tensor(0.5)
    y = rev.Tensor(1)
    z = rev.Tensor(3)
    function = 2*rev.sinh(x)*x + 3*rev.cosh(y) + rev.tanh(z)
    function.backward()

    # Calculated using numpy
    true_value = 6.145391963626209
    true_x_deriv = 2.1698165761938757
    true_y_deriv = 3.525603580931404
    true_z_deriv = 0.009866037165440192

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(x.grad, true_x_deriv, rel_tol=1e-12)
    assert math.isclose(y.grad, true_y_deriv, rel_tol=1e-12)
    assert math.isclose(z.grad, true_z_deriv, rel_tol=1e-12)


def test_case_base_exp():
    """
    Testing exponential function
    """
    x = rev.Tensor(2)
    function = rev.exp(x) + rev.exp(-x)
    function.backward()

    # Calculated using numpy
    true_value = 7.524391382167263
    true_x_deriv = 7.253720815694038

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(x.grad, true_x_deriv, rel_tol=1e-12)


def test_case_base_log():
    """
    Testing logarithm function
    """
    x = rev.Tensor(3)
    function = rev.log(x)**2
    function.backward()

    # Calculated using numpy
    true_value = 1.206948960812582
    true_x_deriv = 0.7324081924454066

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(x.grad, true_x_deriv, rel_tol=1e-12)


def test_log2():
    """
    Testing logarithm base 2 function
    """
    x = rev.Tensor(3)
    function = rev.log2(x) ** 2
    function.backward()

    # Calculated using numpy
    true_value = np.log2(3) ** 2
    true_x_deriv = 2 * np.log2(3) / 3 / np.log(2)

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(x.grad, true_x_deriv, rel_tol=1e-12)


def test_log10():
    """
    Testing logarithm base 10 function
    """
    x = rev.Tensor(3)
    function = rev.log10(x) ** 2
    function.backward()

    # Calculated using numpy
    true_value = np.log10(3) ** 2
    true_x_deriv = 2 * np.log10(3) / 3 / np.log(10)

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(x.grad, true_x_deriv, rel_tol=1e-12)


def test_case_base_root():
    """
    Testing square root and cubic root functions
    """
    x = rev.Tensor(4)
    y = rev.Tensor(8)
    function = rev.sqrt(x) + rev.cbrt(y)
    function.backward()

    # Calculated using numpy
    true_value = 4
    true_x_deriv = 0.25
    true_y_deriv = 0.0833333333333333

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(x.grad, true_x_deriv, rel_tol=1e-12)
    assert math.isclose(y.grad, true_y_deriv, rel_tol=1e-12)


def test_case_base_abs():
    """
    Testing absolute value functin
    """
    x = rev.Tensor(-2)
    function = rev.abs(x)
    function.backward()

    true_value = 2
    true_x_deriv = -1

    assert math.isclose(function.value, true_value, rel_tol=1e-12)
    assert math.isclose(x.grad, true_x_deriv, rel_tol=1e-12)


# ##########################
# #       Unit Tests       #
# ##########################


# #### Tests on the Tensor Class Object ####

def test_children_Tensor():
    x = rev.Tensor(1)
    y = rev.Tensor(2)
    z = rev.Tensor(3)

    function = (x + y) * z
    function.backward()

    # Real children
    child1 = x.children
    child2 = y.children
    child3 = z.children
    child4 = child1[0][1].children
    child5 = child2[0][1].children

    # Leaf nodes
    child6 = child3[0][1].children
    child7 = child4[0][1].children
    child8 = child5[0][1].children

    assert(child6 == [])
    assert(child7 == [])
    assert(child8 == [])

    assert(child1 == child2)
    assert(child3 == child4 == child5)


def test_backward_Tensor():
    """
    Makes sure that the .backward() seed is set correct
    """
    x = rev.Tensor(3)
    assert x.grad_value is None

    x.backward()
    assert x.grad_value == 1

    # make sure you can't call .backward() on a non-scalar
    y = x * [1, 2, 3]
    with pytest.raises(ValueError):
        y.backward()


def test_repr():
    """
    Ensure repr correctly prints the Tensor class
    """
    test_tensor = rev.Tensor(3)
    test_tensor2 = rev.Tensor(2)
    test_tensor2.backward()

    assert test_tensor.__repr__() == 'Tensor(3.0, D(None))'
    assert test_tensor2.__repr__() == 'Tensor(2.0, D(1.0))'


def test_neg():
    """
    Ensure negating works on both value and derivatives
    """
    test_tensor = rev.Tensor(3)
    neg_tensor = test_tensor.__neg__()

    assert neg_tensor.value == -3


def test_add():
    """
    Add two tensors together to get a new tensor
    """
    test_tensor1 = rev.Tensor(3)
    test_tensor2 = rev.Tensor(5)

    new_tensor = test_tensor1 + test_tensor2
    new_tensor.backward()

    assert type(new_tensor) == rev.Tensor
    assert new_tensor.value == 8
    assert test_tensor1.grad == 1.0
    assert test_tensor2.grad == 1.0
    assert test_tensor1.children[0][1] == new_tensor
    assert test_tensor2.children[0][1] == new_tensor


def test_radd():
    """
    Add a non-tensor and a tensor
    """
    t1 = rev.Tensor(3)
    res = 5 + t1
    res.backward()

    assert isinstance(res, rev.Tensor)
    assert res.value == 8
    assert t1.grad == 1.0
    assert t1.children[0][1] == res


def test_sub():
    """
    Subtract a tensor and a tensor
    """
    t1 = rev.Tensor(3)
    t2 = rev.Tensor(1)
    res = t1 - t2
    res.backward()

    assert isinstance(res, rev.Tensor)
    assert res.value == 2
    assert t2.grad == -t2.value
    assert t2.children[0][1].value == t2.__neg__().value


def test_rsub():
    """
    Subtract a non-tensor and a tensor
    """
    t1 = rev.Tensor(3)
    res = 5 - t1
    res.backward()

    assert isinstance(res, rev.Tensor)
    assert res.value == 2
    assert t1.grad == -1
    assert t1.children[0][1].value == t1.__neg__().value


def test_truedriv():
    """
    Divide a tensor and a tensor
    """
    t1 = rev.Tensor(3)
    t2 = rev.Tensor(2)
    res = t1 / t2
    res.backward()

    assert isinstance(res, rev.Tensor)
    assert res.value == 3/2
    assert t1.grad == 1/2
    assert t2.grad == -3/4
    assert t1.children[0][1].value == res.value
    assert t2.children[0][1].value == res.value


def test_rtruedriv():
    """
    Divide a non-tensor and a tensor
    """
    t1 = rev.Tensor(3)
    res = 2 / t1
    res.backward()

    assert isinstance(res, rev.Tensor)
    assert res.value == 2/3
    assert t1.grad == -2/9
    assert t1.children[0][1].value == res.value


def test_pow():
    """
    Take a tensor to the power of a constant
    """
    t1 = rev.Tensor(3)
    res = t1 ** 4
    res.backward()

    assert isinstance(res, rev.Tensor)
    assert res.value == 81
    assert t1.grad == 108
    assert t1.children[0][1].value == res.value


def test_rpow():
    """
    Take a non-tensor to the power of a tensor
    """
    t1 = rev.Tensor(3)
    res = 4 ** t1
    res.backward()

    assert isinstance(res, rev.Tensor)
    assert res.value == 64
    assert t1.grad == (64*np.log(4))
    assert t1.children[0][1].value == res.value


def test_logistic():
    """
    Tests the logistic function
    """
    t1 = rev.Tensor(3)
    res = rev.logistic(t1)

    res.backward()

    expected_value = 1 / (1 + np.exp(-3))
    expected_deriv = np.exp(-3) / (np.exp(-3) + 1) ** 2

    assert math.isclose(res.value, expected_value, rel_tol=1e-12)
    assert math.isclose(t1.grad, expected_deriv, rel_tol=1e-12)


def test_loge():
    """
    Tests the log base e function
    """
    t1 = rev.Tensor(3)
    res = rev.log(t1)

    res.backward()

    expected_value = np.log(t1.value)
    expected_deriv = 1/t1.value

    assert res.value == expected_value
    assert t1.grad == expected_deriv


#### Tests on the Tensor Class Object ####

def test_comparisons():
    """
    Check that comparison operators work for Tensors
    """
    t1 = rev.Tensor(2)
    t2 = rev.Tensor(2)
    t3 = rev.Tensor(1)
    t4 = rev.Tensor(3)

    # test against another Tensor
    assert t1 == t2
    assert t1 <= t2
    assert t1 >= t2
    assert t3 < t1
    assert t3 <= t1
    assert t4 > t1
    assert t4 >= t1
    assert t1 != t3

    # test against a non-Tensor
    assert t1 == 2
    assert t1 <= 2
    assert t1 >= 2
    assert t1 < 3
    assert t1 > 1
    assert t1 != 0


def test_bool():
    """
    Check that you should only be able to call bool() on a singleton Tensor
    """
    t0 = rev.Tensor(0)
    t1 = rev.Tensor(1)
    t2 = rev.Tensor(2)

    assert bool(t0) is False
    assert bool(t1) is True
    assert bool(t2) is True


def test_tensor_build():
    """
    Test Tensor init
    """

    # Check that you can build a Tensor from another Tensor
    t0 = rev.Tensor([1, 2, 3])
    t1 = t0 * t0
    res = t1.sum()
    res.backward()

    assert np.allclose(t0.grad, [2, 4, 6])
    assert t0.children != []
    assert t0.grad_value is not None

    s0 = rev.Tensor(t0)
    assert s0.grad_value is None
    assert s0.children == []

    # Check that you can't build a > 2D tensor
    with pytest.raises(ValueError):
        rev.Tensor([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])


def test_matmul():
    """
    Test matrix multiplication functionality
    """

    # test matrix x matrix
    t0 = rev.Tensor([[1, 2], [3, 4]])
    t1 = rev.Tensor([[5, 6], [7, 8]])
    t2 = rev.Tensor([[9, 10], [11, 12]])
    res = (t0 @ t1 @ t2).sum()

    res.backward()

    assert res.value == 2834
    assert np.allclose(t0.grad, [[233, 317], [233, 317]])
    assert np.allclose(t1.grad, [[76, 92], [114, 138]])
    assert np.allclose(t2.grad, [[62, 62], [72, 72]])

    # test matrix x vector
    t0 = rev.Tensor([[1, 2], [3, 4]])
    t1 = rev.Tensor([5, 6])
    res = (t0 @ t1).sum()

    res.backward()

    assert res.value == 56
    assert np.allclose(t0.grad, [[5, 6], [5, 6]])
    assert np.allclose(t1.grad, [4, 6])

    # test __rmatmul__
    t0 = [[1, 2], [3, 4]]
    t1 = rev.Tensor([5, 6])
    res = (t0 @ t1).sum()

    res.backward()
    assert res.value == 56
    assert np.allclose(t1.grad, [4, 6])


def test_broadcast():
    """
    Test gradients when broadcasting
    """

    t0 = rev.Tensor([[1], [2]])        # a 2 x 1 matrix
    t1 = rev.Tensor([[5, 6], [6, 7]])  # a 2 x 2 matrix
    res = (t0 * t1).sum()

    res.backward()

    assert res.value == 37
    assert np.allclose(t0.grad, [[11], [13]])
    assert np.allclose(t1.grad, [[1, 1], [2, 2]])


def test_reset_grad():
    """
    Test out reset_grad()
    """
    t0 = rev.Tensor(10)
    t1 = rev.Tensor(20)

    res = t0 * t1
    res.backward()

    assert t0.grad == 20
    assert t0.grad_value == 20
    assert t0.children != []

    assert t1.grad == 10
    assert t1.grad_value == 10
    assert t1.children != []

    t0.reset_grad()
    t1.reset_grad()

    assert t0.grad_value is None
    assert t0.children == []

    assert t1.grad_value is None
    assert t1.children == []


def test_aggregates():
    """
    Test out aggregates. Sums, products, means
    """

    # test aggregates for vectors
    t0 = rev.Tensor([3, 5, 7])

    res = t0.sum()
    res.backward()
    assert res.value == 15
    assert np.allclose(t0.grad, [1, 1, 1])

    t0.reset_grad()
    res = t0.mean()
    res.backward()
    assert res.value == 5
    assert np.allclose(t0.grad, [1/3, 1/3, 1/3])

    t0.reset_grad()
    res = t0.prod()
    res.backward()
    assert res.value == 105
    assert np.allclose(t0.grad, [35, 21, 15])

    # test aggregates for scalars
    t1 = rev.Tensor(5)

    res = t1.sum()
    res.backward()
    assert res.value == 5
    assert t1.grad == 1

    t1.reset_grad()
    res = t1.mean()
    res.backward()
    assert res.value == 5
    assert t1.grad == 1

    t1.reset_grad()
    res = t1.prod()
    res.backward()
    assert res.value == 5
    assert t1.grad == 1


def test_no_grad():
    """
    Check that the computational graph is not built while no_grad() is active
    """
    t0 = rev.Tensor(5)
    t1 = t0 * t0
    assert t0.children != []

    with rev.no_grad():
        t0 = rev.Tensor(5)
        t1 = t0 * t0
        assert t0.children == []


def test_get_set_item():
    """
    Test getting items in a Tensor
    """
    t0 = rev.Tensor([[1, 2, 3, 4], [5, 6, 7, 8]])
    s1 = t0[0, :]
    s2 = t0[1, :]
    res = (s1 * s2).sum()

    res.backward()
    assert np.allclose(s1.value, [1, 2, 3, 4])
    assert np.allclose(s2.value, [5, 6, 7, 8])

    assert res.value == 70
    assert np.allclose(s1.grad, [5, 6, 7, 8])
    assert np.allclose(s2.grad, [1, 2, 3, 4])
    assert np.allclose(t0.grad, [[5, 6, 7, 8], [1, 2, 3, 4]])


def test_set_item():
    """
    Test setting items in a Tensor
    """
    t0 = rev.Tensor([[1, 1, 1], [1, 1, 1]])
    s1 = rev.Tensor([1000, 2])
    t0[0, 1:3] = s1

    res = t0.prod()
    res.backward()

    assert res.value == 2000
    assert np.allclose(t0.grad, [[2000, 2, 1000], [2000, 2000, 2000]])
    assert np.allclose(s1.grad, [2, 1000])



import numpy as np
from collections import defaultdict

# we use this to get rid of disruptive runtime warnings
runtime_warning_filter = np.testing.suppress_warnings()
runtime_warning_filter.filter(RuntimeWarning)


class MultivariateDerivative:
    """
    Multivariate Derivative is a class called automatically by the Variable
    or Tensor classes to hold derivative information.

    Parameters
    ----------
    variables : dict
        Store the name and value of the derivative object allowing for multiple
        variable keys

    Attributes
    ----------
    variables : dict
        Internal storage of the named derivative values in dictionary format
    """

    def __init__(self, variables=None):
        variables = variables or {}
        self.variables = {k: np.array(v) for k, v in variables.items()}

    def __repr__(self):
        """
        Create a descriptive object for printing, D noting that the object is a
        derivative

        Returns
        -------
        str
            a string representing the derivative w.r.t all of the variables
        """
        values = ', '.join(f'{k}={v}' for k, v in self.variables.items())
        return f'D({values})'

    def copy(self):
        """
        Given an existing MultivariateDerivative object (self) create a new one
        as a copy

        Returns
        -------
        MultivariateDerivative
            a new MultivariateDerivative object
        """
        return MultivariateDerivative(self.variables.copy())

    def __getitem__(self, key):
        """
        Access the derivative of one of the variables in the object.
        instance.variables.keys() will reveal the existing keys within the
        instance

        Returns
        -------
        dict value
            Value of the dictionary of derivatives at the specified key
        """
        return self.variables[key]

    # the only way to combine derivatives is with addition or multiplication
    def __add__(self, other):
        """
        Adds this and another MultivariateDerivative object together and
        produces a new MultivariateDerivative object whose variable set is the
        union of the two individuals.

        Returns
        -------
        MultivariateDerivative
            a new MultivariateDerivative object
        """
        var_set = set(self.variables.keys()) | set(other.variables.keys())
        result = {}
        for v in var_set:
            a, b = self.variables.get(v, 0), other.variables.get(v, 0)
            a, b = self._broadcast(a, b)
            result[v] = a + b
        return MultivariateDerivative(result)

    def mul(self, multiplier):
        """
        Multiplies all of the values in this MultivariateDerivative by a
        scalar (i.e. non-Tensor and non-MultivariateDerivative) object and
        return a new MultivariateDerivative.

        Returns
        -------
        MultivariateDerivative
            a new MultivariateDerivative object
        """
        result = {}
        for k, v in self.variables.items():
            a, b = self._broadcast(multiplier, v)
            result[k] = a * b
        return MultivariateDerivative(result)

    def get_idx(self, i):
        """
        Returns a copy of the MultivariateDerivative with the desired index
        `i` from each variable selected

        Returns
        -------
        MultivariateDerivative
            a new MultivariateDerivative object
        """
        result = {k: v[i] for k, v in self.variables.items()}
        return MultivariateDerivative(result)

    def set_idx(self, i, other, tensor_value):
        """
        Modifies the MultivariateDerivative by updating with the desired index
        `i` in each variable. There cannot be any variables in `other` that
        are not already in the current variable set

        Returns
        -------
        None
        """
        for k, v in self.variables.items():
            if k not in other.variables:
                self.variables[k][i] *= 0

        for k, v in other.variables.items():
            if k not in self.variables:
                self.variables[k] = np.zeros(tensor_value.shape)
            self.variables[k][i] = other.variables[k]

    def _fix_shape(self, value):
        """
        If a vector Tensor and a scalar variable are added, then it's possible
        that the MultivariateDerivative that results won't have the right shape.
        This fixes any shapes based on the shape of the Tensor by broadcasting

        Returns
        -------
        None
        """
        for k, v in self.variables.items():
            if len(v.shape) < len(value.shape):
                a, b = self._broadcast(value, v)
                self.variables[k] = np.zeros(a.shape, dtype=b.dtype) + b

    def _broadcast(self, v1, v2):
        """
        Given two np.arrays, this function figures out the appropriate indexing
        for both variables which would allow the two operands to broadcast.
        """
        v1, v2 = np.array(v1), np.array(v2)
        if len(v1.shape) < len(v2.shape):
            idx = tuple(slice(None) for i in range(len(v1.shape)))
            idx = idx + (None,) * (len(v2.shape) - len(v1.shape))
            return v1[idx], v2
        elif len(v1.shape) > len(v2.shape):
            idx = tuple(slice(None) for i in range(len(v2.shape)))
            idx = idx + (None,) * (len(v1.shape) - len(v2.shape))
            return v1, v2[idx]
        else:
            return v1, v2


class Tensor:
    """
    Tensor (constant value) is the base class of the Variable and can be used to
    represent constant values where the derivative is not required for automatic
    computation during automatic differentiation. Derivative by default is set
    to an empty dictionary.

    Parameters
    ----------
    value : np.array
        Store the value of the constant

    d : MultivariateDerivative
        Stores the value of the derivative. The default is an empty
        MultivariateDerivative, but it can be manually set to a specific value

    Attributes
    ----------
    value : np.array
        Internal storage of the tensor value

    d : MultivariateDerivative
        Internal storage of the tensor derivative
    """
    def __init__(self, value, d=None):
        # If existing Tensor check for whether a derivative value was set
        if isinstance(value, Tensor):
            value, d = value.value, value.d.copy()

        self.value = np.array(value)
        self.d = d or MultivariateDerivative()
        self.d._fix_shape(self.value)

    @staticmethod
    def get_value_and_deriv(other):
        if isinstance(other, Tensor):
            return other.value, other.d
        return other, MultivariateDerivative()

    def sum(self):
        if self.shape == ():
            return Tensor(self)
        return sum(self, Tensor(0))

    def norm(self, d=2):
        if self.shape == ():
            return Tensor(self)
        return power(sum(x ** d for x in self), 1/d)

    def all(self):
        return bool(np.all(self.value))

    def any(self):
        return bool(np.any(self.value))

    def __len__(self):
        return len(self.value)

    def __iter__(self):
        size = len(self)
        return (self[i] for i in range(size))

    @property
    def shape(self):
        return self.value.shape

    # Comparisons work like they do in numpy. Derivative information is ignored
    # numpy arrays are returned by comparisons
    def __lt__(self, other):
        other_v, _ = Tensor.get_value_and_deriv(other)
        return self.value < other_v

    def __gt__(self, other):
        other_v, _ = Tensor.get_value_and_deriv(other)
        return self.value > other_v

    def __le__(self, other):
        other_v, _ = Tensor.get_value_and_deriv(other)
        return self.value <= other_v

    def __ge__(self, other):
        other_v, _ = Tensor.get_value_and_deriv(other)
        return self.value >= other_v

    def __eq__(self, other):
        other_v, _ = Tensor.get_value_and_deriv(other)
        return self.value == other_v

    def __ne__(self, other):
        other_v, _ = Tensor.get_value_and_deriv(other)
        return self.value != other_v

    def __bool__(self):
        return bool(self.value)

    def __repr__(self):
        return f'Tensor({self.value}, {self.d})'

    def __neg__(self):
        return Tensor(-self.value, self.d.mul(-1))

    def __add__(self, other):
        other_v, other_d = Tensor.get_value_and_deriv(other)
        return Tensor(self.value + other_v, self.d + other_d)

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        other_v, other_d = Tensor.get_value_and_deriv(other)
        return Tensor(self.value - other_v, self.d + other_d.mul(-1))

    def __rsub__(self, other):
        other_v, other_d = Tensor.get_value_and_deriv(other)
        return Tensor(other_v - self.value, other_d + self.d.mul(-1))

    def __mul__(self, other):
        other_v, other_d = Tensor.get_value_and_deriv(other)
        prod_rule = self.d.mul(other_v) + other_d.mul(self.value)
        return Tensor(self.value * other_v, prod_rule)

    def __rmul__(self, other):
        return self * other

    def __truediv__(self, other):
        other_v, other_d = Tensor.get_value_and_deriv(other)
        quot_rule = self.d.mul(other_v) + other_d.mul(-self.value)
        quot_rule = quot_rule.mul(1 / (other_v ** 2))
        return Tensor(self.value / other_v, quot_rule)

    def __rtruediv__(self, other):
        other_v, other_d = Tensor.get_value_and_deriv(other)
        quot_rule = other_d.mul(self.value) + self.d.mul(-other_v)
        quot_rule = quot_rule.mul(1 / (self.value ** 2))
        return Tensor(other_v / self.value, quot_rule)

    def __pow__(self, other):
        return power(self, other)

    def __rpow__(self, other):
        return power(other, self)

    # TODO: matrix operators like dot prod, matrix mult, inverse, tranpose, etc
    def __getitem__(self, i):
        return Tensor(self.value[i], self.d.get_idx(i))

    def __setitem__(self, i, other):
        other_v, other_d = Tensor.get_value_and_deriv(other)
        self.d.set_idx(i, other_d, self.value)
        self.value[i] = other_v


class Variable(Tensor):
    """
    Variable (a variable value for the purposes of differentiation). A variable
    is used to represent any value which the user wishes to include in the
    partial derivative outputs and is built on top of the Tensor base class.

    As a variable a name parameter is required (example: x, y or z).

    Parameters
    ----------
    name: str
        Store the name of the variable

    value : np.array
        Store the value of the variable

    d : MultivariateDerivative
        Store the value of the derivative, default is the identity matrix
        which will result in the computation of the Jacobian

    Attributes
    ----------
    name: str
        Internal storage of the variable name

    value : np.array
        Internal storage of the variable value

    d : MultivariateDerivative
        Internal storage of the variable derivative
    """
    def __init__(self, name, value):
        self.name = name
        self.value = np.array(value)

        if len(self.value.shape) >= 2:
            raise ValueError('Variable does not support dimensions >= 2')

        if self.value.shape == (0,):
            raise ValueError('Cannot have an empty value for Variable')

        if self.value.shape == ():
            self.d = MultivariateDerivative({
                name: np.array(1)
            })
        else:
            self.d = MultivariateDerivative({
                name: np.eye(len(self.value))
            })


# Elementary operations of a single variable. They all use chain rule.
@runtime_warning_filter
def _elementary_op(obj, fn, deriv_fn):
    """
    A generic framework to allow for the chain rule of other elementary
    functions taken from the numpy module.

    Parameters
    ----------
    obj : Class
        The Variable or Tensor which the elementary function is being
        differentiated at

    fn : np.function
        The elementary function from the numpy module

    deriv_fun:  np.function
        The prespecified derivative of the given numpy elementary function


    Returns
    -------
    Tensor: class
        Tensor object which contains the resulting value and result from the
        chain rule (new derivative)
    """
    v, d = Tensor.get_value_and_deriv(obj)
    chain_rule = d.mul(deriv_fn(v))
    return Tensor(fn(v), chain_rule)


def sin(tensor):
    """
    pyad sin - computes the sine function
        sine differentiates to cosine

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the sin function is
        being differentiated at

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    return _elementary_op(tensor, np.sin, np.cos)


def cos(tensor):
    """
    pyad cos - computes the cosine function
        cosine differentiates to minus sine

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the cosine function is
        being differentiated at

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    return _elementary_op(tensor, np.cos, lambda x: -np.sin(x))


def tan(tensor):
    """
    pyad tan - computes the tangent function
        The tangent of x differentiates to sec^2(x) (1/cos^2(x))

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the tangent function is
        being differentiated at

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    return _elementary_op(tensor, np.tan, lambda x: 1 / (np.cos(x) ** 2))


def arcsin(tensor):
    """
    pyad arcsin - computes the arcsine function
        The arcsine of x differentiates to 1/√(1-x^2)

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the arcsine function is
        being differentiated at

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    return _elementary_op(tensor, np.arcsin, lambda x: 1 / np.sqrt(1 - x ** 2))


def arccos(tensor):
    """
    pyad arccos - computes the arccosine function
        The arccosine of x differentiates to -1/√(1-x^2)

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the arccosine function is
        being differentiated at

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    return _elementary_op(tensor, np.arccos, lambda x: -1 / np.sqrt(1 - x ** 2))


def arctan(tensor):
    """
    pyad arctan - computes the arctangent function
        The arctangent of x differentiates to 1/(1+x^2)

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the arctangent function is
        being differentiated at

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    return _elementary_op(tensor, np.arctan, lambda x: 1 / (1 + x ** 2))


def sinh(tensor):
    '''
    pyad sinh - computes the hyperbolic sine function
        The sinh of x differentiates to cosh(x)

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the sinh function is
        being differentiated at

    Returns
    -------
    Tensor: class
        Calls the _elementary_op function and returns the resulting Tensor
    '''
    return _elementary_op(tensor, np.sinh, np.cosh)


def cosh(tensor):
    '''
    pyad cosh - computes the hyperbolic cosine function
        The cosh of x differentiates to sinh(x)

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the cosh function is
        being differentiated at

    Returns
    -------
    Tensor: class
        Calls the _elementary_op function and returns the resulting Tensor
    '''
    return _elementary_op(tensor, np.cosh, np.sinh)


def tanh(tensor):
    '''
    pyad tanh - computes the hyperbolic tangent function
        The tanh of x differentiates to 1/cosh^2(x)

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the cosh function is
        being differentiated at

    Returns
    -------
    Tensor: class
        Calls the _elementary_op function and returns the resulting Tensor
    '''
    return _elementary_op(tensor, np.tanh, lambda x: 1/(np.cosh(x)**2))


def abs(tensor):
    """
    pyad abs - computes the absolute value function
        For simplicity, we have defined the absolute value derviative to be the
        sign of the argument:
            if x > 0:  return 1
            if x == 0: return 0
            if x < 0:  return -1

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the absolute value function
        is being differentiated at

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    return _elementary_op(tensor, np.abs, np.sign)


def exp(tensor):
    """
    pyad exp - computes the exponential function, e^x
        e^x differentiates to e^x

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the exponential function is
        being differentiated at

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    return _elementary_op(tensor, np.exp, np.exp)


def log(tensor, base=np.e):
    """
    pyad log - computes the natural logarithm function
        The natural logarithm of x differentiates to 1/x

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the natural logarithm is
        being differentiated at

    base: float | Tensor
        The base of the logarithm. Defaults to `np.e`. Can be a Tensor

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    if base == np.e:
        return _elementary_op(tensor, np.log, lambda x: 1 / x)
    return log(tensor) / log(base)


def log2(tensor):
    """
    pyad log2 - a wrapper of log to computes logarithm base 2

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the function is
        being differentiated at

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    return log(tensor, base=2)


def log10(tensor):
    """
    pyad log10 - a wrapper of log to computes logarithm base 10

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the function is
        being differentiated at

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    return log(tensor, base=10)


def sqrt(tensor):
    """
    pyad sqrt - computes the square root function
        The square root of x differentiates to 1/(2√x)

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the square root is
        being differentiated at

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    return _elementary_op(tensor, np.sqrt, lambda x: 1 / (2 * np.sqrt(x)))


def cbrt(tensor):
    """
    pyad cbrt - computes the cube root function
        The cube root of x differentiates to 1/(3x^(2/3))

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the cube root is
        being differentiated at

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    return _elementary_op(tensor, np.cbrt, lambda x: 1 / (3 * np.power(x, 2/3)))


@runtime_warning_filter
def power(base, exp):
    """
    pyad power - computes the power function
        The power function is differentiated using the generalized power rule:
            d/dx[a^b] = d/dx[exp(ln(a) * b)] = a^b * (b' * ln(a) + a'/a * b)

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the power function is
        being differentiated at

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    base_v, base_d = Tensor.get_value_and_deriv(base)
    exp_v, exp_d = Tensor.get_value_and_deriv(exp)

    result = base_v ** exp_v
    a = base_d.mul(exp_v * base_v ** (exp_v - 1.0))
    b = exp_d.mul(result * np.log(base_v))
    return Tensor(result, a + b)


def logistic(tensor):
    """
    pyad logistic - computes the logistic function: 1 / (1 + exp(-x))

    Parameters
    ----------
    tensor : class
        The value of the variable or constant which the logistic function is
        being differentiated at

    Returns
    -------
    Tensor: class
        Applies chain rule as appropriate and returns the resulting Tensor
    """
    return 1 / (1 + exp(-tensor))


# functions for vector operatoions
def stack(tensors):
    """
    pyad stack - combines multiple pyad Tensors into a single tensor by stacking

    Parameters
    ----------
    tensors : List[Tensor]
        The list of tensors that stack

    Returns
    -------
    Tensor: class
        The resulting derivative is computed by stacking the individual components
    """
    if len(tensors) == 0:
        raise ValueError('need at least one tensor to stack')

    values = []
    derivs = []
    var_set = set()
    for t in tensors:
        v, d = Tensor.get_value_and_deriv(t)
        values.append(v)
        derivs.append(d)
        var_set |= set(d.variables.keys())

    variables = defaultdict(list)
    for k in var_set:
        d_vals = [d.variables[k] for d in derivs if k in d.variables]
        if not all(v.shape == d_vals[0].shape for v in d_vals):
            raise ValueError('all input tensors must have the same shape')

        for d in derivs:
            if k not in d.variables:
                variables[k].append(np.zeros(d_vals[0].shape))
            else:
                variables[k].append(d.variables[k])

        variables[k] = np.array(variables[k])

    stacked_d = MultivariateDerivative(dict(variables))
    return Tensor(np.stack(values), stacked_d)


# wrappers around Tensor and Variable constructors
def tensor(*args, **kwargs):
    """
    tensor - a wrapper for the Tensor class constructor
    """
    return Tensor(*args, **kwargs)


def var(*args, **kwargs):
    """
    var - a wrapper for the Variable class constructor
    """
    return Variable(*args, **kwargs)



from setuptools import setup, find_packages

setup(
    name='pyad207',
    description='A simple package for auto-differentiation.',
    version='1.0.0',
    packages=find_packages(),
    url='https://github.com/fire-breathing-rubber-lemons/cs207-FinalProject',
    author='CS207 Project Group 10',
    python_requires='>=3.6',
    install_requires=[
        'numpy>=1.16.3',
        'networkx>=2.4',
        'matplotlib>=3.1.0'
    ],
)
